import ts from "typescript";
import SchemaTypeBuilder from "./SchemaTypeBuilder.js";
import QueryTypeBuilder from "./QueryTypeBuilder.js";
import { normalize, replaceRange } from "../util.js";
import fs from "fs";
/**
 * The file visitor is ephemeral. Created each time we visit a file.
 */
export default class FileVisitor {
    options;
    schemaCache;
    dag;
    sourceFile;
    sqlTemplates = [];
    schemaTemplates = [];
    constructor(options, schemaCache, dag, sourceFile) {
        this.options = options;
        this.schemaCache = schemaCache;
        this.dag = dag;
        this.sourceFile = sourceFile;
    }
    visitAll(checker) {
        this.dag.orphan(this.sourceFile.fileName);
        this.collectAllNodes(this.sourceFile, checker);
        const schemaTypeBuilder = new SchemaTypeBuilder(this.options, this.schemaCache, this.sourceFile);
        const schemaFixes = schemaTypeBuilder.buildResidentTypes(this.schemaTemplates);
        const queryFixes = new QueryTypeBuilder(this.schemaCache, this.dag, this.sourceFile).buildQueryTypes(this.sqlTemplates, checker);
        this.applyFixes(schemaFixes.concat(queryFixes));
    }
    visitSchemaDefs(checker) {
        this.collectSchemaNodes(this.sourceFile, checker);
        const schemaFixes = new SchemaTypeBuilder(this.options, this.schemaCache, this.sourceFile).buildResidentTypes(this.schemaTemplates);
        this.applyFixes(schemaFixes);
    }
    visitQueryDefs(checker) {
        this.collectQueryNodes(this.sourceFile, checker);
        const queryFixes = new QueryTypeBuilder(this.schemaCache, this.dag, this.sourceFile).buildQueryTypes(this.sqlTemplates, checker);
        this.applyFixes(queryFixes);
    }
    applyFixes(fixes) {
        if (fixes.length == 0) {
            return;
        }
        let offset = 0;
        console.log("Applying fix to " + this.sourceFile.fileName);
        for (const fix of fixes) {
            switch (fix._tag) {
                case "InlineFix": {
                    const { range, replacement } = fix;
                    const text = this.sourceFile.getFullText();
                    const updatedText = replaceRange(text, range[0], range[1], replacement);
                    // if replacement is _longer_ than what was replaced then all other replacements shift further away
                    // if replacement is _shorter_ then they shift closer
                    offset += replacement.length - (range[1] - range[0]);
                    fs.writeFileSync(this.sourceFile.fileName, updatedText);
                    break;
                }
                case "CompanionFileFix": {
                    // write the companion file
                    this.#writeCompanionFile(fix);
                }
            }
        }
    }
    // only write it if it differs from what is on disk after normalization
    #writeCompanionFile(fix) {
        let prefix = fix.placeAfter ?? "";
        if (fs.existsSync(fix.path)) {
            let existing = fs.readFileSync(fix.path, "utf-8");
            let placeAfterEnd = 0;
            if (fix.placeAfter != null) {
                const placeAfterStart = existing.indexOf(fix.placeAfter);
                if (placeAfterStart != -1) {
                    placeAfterEnd = placeAfterStart + fix.placeAfter.length;
                    prefix = existing.substring(0, placeAfterEnd);
                    existing = existing.substring(placeAfterEnd);
                }
            }
            if (normalize(existing) == normalize(fix.content)) {
                console.log(`No difference for companion file ${fix.path}`);
                return;
            }
        }
        fs.writeFileSync(fix.path, prefix + fix.content);
    }
    collectAllNodes(node, checker) {
        if (!ts.isTaggedTemplateExpression(node)) {
            ts.forEachChild(node, (node) => this.collectAllNodes(node, checker));
            return;
        }
        const tagName = node.tag.getText();
        if (tagName.endsWith(".sql")) {
            this.sqlTemplates.push(node);
        }
        else if (tagName.endsWith("schema")) {
            this.schemaTemplates.push(node);
        }
    }
    collectSchemaNodes(node, checker) {
        if (!ts.isTaggedTemplateExpression(node)) {
            ts.forEachChild(node, (node) => this.collectSchemaNodes(node, checker));
            return;
        }
        const tagName = node.tag.getText();
        if (tagName.endsWith("schema")) {
            this.schemaTemplates.push(node);
        }
    }
    collectQueryNodes(node, checker) {
        if (!ts.isTaggedTemplateExpression(node)) {
            ts.forEachChild(node, (node) => this.collectQueryNodes(node, checker));
            return;
        }
        const tagName = node.tag.getText();
        if (tagName.endsWith(".sql")) {
            this.sqlTemplates.push(node);
        }
    }
}
//# sourceMappingURL=FileVisitor.js.map